import logging
from logging import StreamHandler
import sys

unicode_type = str
basestring_type = str
_TO_UNICODE_TYPES = (unicode_type, type(None))


def to_unicode(value):
    """Converts a string argument to a unicode string.
    If the argument is already a unicode string or None, it is returned
    unchanged.  Otherwise it must be a byte string and is decoded as utf8.
    """
    if isinstance(value, _TO_UNICODE_TYPES):
        return value
    if not isinstance(value, bytes):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.decode("utf-8")


# to_unicode was previously named _unicode not because it was private,
# but to avoid conflicts with the built-in unicode() function/type
_unicode = to_unicode

try:
    import colorama
except ImportError:
    colorama = None

try:
    import curses
except ImportError:
    curses = None


def _stderr_supports_color():
    try:
        if hasattr(sys.stderr, 'isatty') and sys.stderr.isatty():
            if curses:
                curses.setupterm()
                if curses.tigetnum("colors") > 0:
                    return True
            elif colorama:
                if sys.stderr is getattr(colorama.initialise, 'wrapped_stderr',
                                         object()):
                    return True
    except Exception:
        # Very broad exception handling because it's always better to
        # fall back to non-colored logs than to break at startup.
        pass
    return False


def _safe_unicode(s):
    try:
        return _unicode(s)
    except UnicodeDecodeError:
        return repr(s)


class LogFormatter(logging.Formatter):
    """Log formatter used from Tornado.
    """
    DEFAULT_FORMAT = '%(color)s[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s'
    DEFAULT_DATE_FORMAT = '%y%m%d %H:%M:%S'
    DEFAULT_COLORS = {
        logging.DEBUG: 4,  # Blue
        logging.INFO: 2,  # Green
        logging.WARNING: 3,  # Yellow
        logging.ERROR: 1,  # Red
    }

    def __init__(self, fmt=DEFAULT_FORMAT, datefmt=DEFAULT_DATE_FORMAT, style='%', color=True, colors=DEFAULT_COLORS):
        """
        :arg bool color: Enables color support.
        :arg string fmt: Log message format.
          It will be applied to the attributes dict of log records. The
          text between ``%(color)s`` and ``%(end_color)s`` will be colored
          depending on the level if color support is on.
        :arg dict colors: color mappings from logging level to terminal color
          code
        :arg string datefmt: Datetime format.
          Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``.
        """
        logging.Formatter.__init__(self, datefmt=datefmt)
        self._fmt = fmt

        self._colors = {}
        if color and _stderr_supports_color():
            if curses is not None:
                # The curses module has some str/bytes confusion in
                # python3.  Until version 3.2.3, most methods return
                # bytes, but only accept strings.  In addition, we want to
                # output these strings with the logging module, which
                # works with unicode strings.  The explicit calls to
                # unicode() below are harmless in python2 but will do the
                # right conversion in python 3.
                fg_color = (curses.tigetstr("setaf") or
                            curses.tigetstr("setf") or "")
                if (3, 0) < sys.version_info < (3, 2, 3):
                    fg_color = unicode_type(fg_color, "ascii")

                for levelno, code in colors.items():
                    self._colors[levelno] = unicode_type(curses.tparm(fg_color, code), "ascii")
                self._normal = unicode_type(curses.tigetstr("sgr0"), "ascii")
            else:
                # If curses is not present (currently we'll only get here for
                # colorama on windows), assume hard-coded ANSI color codes.
                for levelno, code in colors.items():
                    self._colors[levelno] = '\033[2;3%dm' % code
                self._normal = '\033[0m'
        else:
            self._normal = ''

    def format(self, record):
        try:
            message = record.getMessage()
            assert isinstance(message, basestring_type)  # guaranteed by logging
            # Encoding notes:  The logging module prefers to work with character
            # strings, but only enforces that log messages are instances of
            # basestring.  In python 2, non-ascii bytestrings will make
            # their way through the logging framework until they blow up with
            # an unhelpful decoding error (with this formatter it happens
            # when we attach the prefix, but there are other opportunities for
            # exceptions further along in the framework).
            #
            # If a byte string makes it this far, convert it to unicode to
            # ensure it will make it out to the logs.  Use repr() as a fallback
            # to ensure that all byte strings can be converted successfully,
            # but don't do it by default so we don't add extra quotes to ascii
            # bytestrings.  This is a bit of a hacky place to do this, but
            # it's worth it since the encoding errors that would otherwise
            # result are so useless (and tornado is fond of using utf8-encoded
            # byte strings wherever possible).
            record.message = _safe_unicode(message)
        except Exception as e:
            record.message = "Bad message (%r): %r" % (e, record.__dict__)

        record.asctime = self.formatTime(record, self.datefmt)

        if record.levelno in self._colors:
            record.color = self._colors[record.levelno]
            record.end_color = self._normal
        else:
            record.color = record.end_color = ''

        formatted = self._fmt % record.__dict__

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            # exc_text contains multiple lines.  We need to _safe_unicode
            # each line separately so that non-utf8 bytes don't cause
            # all the newlines to turn into '\n'.
            lines = [formatted.rstrip()]
            lines.extend(_safe_unicode(ln) for ln in record.exc_text.split('\n'))
            formatted = '\n'.join(lines)
        return formatted.replace("\n", "\n    ")


def enable_pretty_logging(logger, handler=None, color=True, level=logging.WARNING):
    """Turns on formatted logging output as configured.
    """
    if not handler:
        channel = StreamHandler()
    else:
        channel = handler
    channel.setFormatter(LogFormatter(color=color))
    logger.addHandler(channel)
    logger.setLevel(level)
    return logger



from werkzeug.wrappers import Request as RequestBase
from werkzeug.wrappers import Response as ResponseBase
from werkzeug.wrappers.json import JSONMixin as _JSONMixin


class Request(RequestBase, _JSONMixin):
    pass


class Response(ResponseBase):
    pass



from werkzeug.wrappers import BaseResponse
from werkzeug.routing import Map, Rule, MapAdapter
from werkzeug.exceptions import HTTPException, InternalServerError, NotFound
from werkzeug.datastructures import Headers
from werkzeug.serving import run_simple
from madara.blueprints import Blueprint
from madara.wrappers import Request, Response
from madara.utils import jsonify, reraise, _endpoint_from_view_func, import_string, load_config
from madara.compat import text_type, string_types
from madara.log import enable_pretty_logging
import sys
import logging
import traceback


class Madara(object):

    def __init__(self, **kwargs):
        self.config = load_config(kwargs)
        self.bootstrap()
        self.url_map: Map = Map()
        self.url_rule_class = Rule
        self.endpoint_map: dict = {}
        self.blueprints: dict = {}
        self.response_class = Response
        self._middleware_chain = None
        self._view_middleware = []
        self._exception_middleware = []
        self.load_middleware()

    def bootstrap(self):
        if self.config.get("debug"):
            self.logger = enable_pretty_logging(logger=logging.getLogger("madara"), level=logging.DEBUG)
        else:
            self.logger = enable_pretty_logging(logger=logging.getLogger("madara"), level=logging.WARNING)

    def load_middleware(self):
        middlewares = self.config.get("middlewares", [])
        handler = self.dispatch_request
        for md in reversed(middlewares):
            mw = import_string(md)
            mw_instance = mw(handler, self)
            if hasattr(mw_instance, 'process_view'):
                self._view_middleware.insert(0, mw_instance.process_view)

            if hasattr(mw_instance, 'process_exception'):
                self._exception_middleware.append(mw_instance.process_exception)
            handler = mw_instance

        self._middleware_chain = handler

    def add_url_rule(self, pattern: str, endpoint=None, view_func=None, provide_automatic_options=None, **options):
        if endpoint is None:
            endpoint = _endpoint_from_view_func(view_func)
        options["endpoint"] = endpoint
        methods = options.pop("methods", None)

        # if the methods are not given and the view_func object knows its
        # methods we can use that instead.  If neither exists, we go with
        # a tuple of only ``GET`` as default.
        if methods is None:
            methods = getattr(view_func, "methods", None) or ("GET",)
        if isinstance(methods, string_types):
            raise TypeError(
                "Allowed methods have to be iterables of strings, "
                'for example: @app.route(..., methods=["POST"])'
            )
        methods = set(item.upper() for item in methods)

        # Methods that should always be added
        required_methods = set(getattr(view_func, "required_methods", ()))

        if provide_automatic_options is None:
            provide_automatic_options = getattr(
                view_func, "provide_automatic_options", None
            )

        if provide_automatic_options is None:
            if "OPTIONS" not in methods:
                provide_automatic_options = True
                required_methods.add("OPTIONS")
            else:
                provide_automatic_options = False

        # Add the required methods now.
        methods |= required_methods

        rule = self.url_rule_class(pattern, methods=methods, **options)
        rule.provide_automatic_options = provide_automatic_options

        self.url_map.add(rule)
        if view_func is not None:
            old_func = self.endpoint_map.get(endpoint)
            if old_func is not None and old_func != view_func:
                raise AssertionError(
                    "View function mapping is overwriting an "
                    "existing endpoint function: %s" % endpoint
                )
            self.endpoint_map[endpoint] = view_func

    def route(self, pattern: str, **options):
        def decorator(func):
            endpoint = options.pop("endpoint", None)
            self.add_url_rule(pattern, endpoint, func, **options)
            return func

        return decorator

    def register_blueprint(self, blueprint: Blueprint, **options):

        if blueprint.name in self.blueprints:
            assert self.blueprints[blueprint.name] is blueprint, (
                "A name collision occurred between blueprints %r and %r. Both"
                ' share the same name "%s". Blueprints that are created on the'
                " fly need unique names."
                % (blueprint, self.blueprints[blueprint.name], blueprint.name)
            )
        else:
            self.blueprints[blueprint.name] = blueprint
        blueprint.register(self, options)

    def dispatch_request(self, request):
        adapter: MapAdapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, view_kwargs = adapter.match()
            endpoint_func = self.endpoint_map.get(endpoint, None)
            if not endpoint_func:
                raise NotFound()
            rv = self.process_view_by_middleware(request, endpoint_func, view_kwargs)
            if rv is None:
                rv = endpoint_func(request, **view_kwargs)
            return self.make_response(request, rv)
        except HTTPException as e:
            return e
        except Exception as e:
            if not self._exception_middleware:
                # if no exception process middleware log the traceback.
                self.logger.error(traceback.format_exc())
            try:
                rv = self.process_exception_by_middleware(request, e)
                if rv is None:
                    return InternalServerError(original_exception=e)
                return self.make_response(request, rv)
            except Exception as e:
                # if exception process middleware raise a exception, log the traceback and return an InternalServerError.
                self.logger.error(traceback.format_exc())
                return InternalServerError(original_exception=e)

    def process_view_by_middleware(self, request, callback, callback_kwargs):
        """
        Pass the request and view_func、view_kwargs to the view middleware.
        Middleware process_view should return either None or a response.
        If it returns None, will continue processing this request, executing any other process_view() middleware and, then, the appropriate view_func.
        """
        for middleware_method in self._view_middleware:
            response = middleware_method(request, callback, callback_kwargs)
            if response:
                return response
        return None

    def process_exception_by_middleware(self, request, exception):
        """
        Pass the exception to the exception middleware. If no middleware
        return a response for this exception, return None.
        """
        for middleware_method in self._exception_middleware:
            response = middleware_method(request, exception)
            if response:
                return response
        return None

    def make_response(self, request, rv):

        status = headers = None

        # unpack tuple returns
        if isinstance(rv, tuple):
            len_rv = len(rv)

            # a 3-tuple is unpacked directly
            if len_rv == 3:
                rv, status, headers = rv
            # decide if a 2-tuple has status or headers
            elif len_rv == 2:
                if isinstance(rv[1], (Headers, dict, tuple, list)):
                    rv, headers = rv
                else:
                    rv, status = rv
            # other sized tuples are not allowed
            else:
                raise TypeError(
                    "The view function did not return a valid response tuple."
                    " The tuple must have the form (body, status, headers),"
                    " (body, status), or (body, headers)."
                )

        # the body must not be None
        if rv is None:
            raise TypeError(
                "The view function did not return a valid response. The"
                " function either returned None or ended without a return"
                " statement."
            )

        # make sure the body is an instance of the response class
        if not isinstance(rv, self.response_class):
            if isinstance(rv, (text_type, bytes, bytearray)):
                # let the response class set the status and headers instead of
                # waiting to do it manually, so that the class can handle any
                # special logic
                rv = self.response_class(rv, status=status, headers=headers)
                status = headers = None
            elif isinstance(rv, dict):
                rv = jsonify(rv)
            elif isinstance(rv, BaseResponse) or callable(rv):
                # evaluate a WSGI callable, or coerce a different response
                # class to the correct type
                try:
                    rv = self.response_class.force_type(rv, request.environ)
                except TypeError as e:
                    new_error = TypeError(
                        "{e}\nThe view function did not return a valid"
                        " response. The return type must be a string, dict, tuple,"
                        " Response instance, or WSGI callable, but it was a"
                        " {rv.__class__.__name__}.".format(e=e, rv=rv)
                    )
                    reraise(TypeError, new_error, sys.exc_info()[2])
            else:
                raise TypeError(
                    "The view function did not return a valid"
                    " response. The return type must be a string, dict, tuple,"
                    " Response instance, or WSGI callable, but it was a"
                    " {rv.__class__.__name__}.".format(rv=rv)
                )

        # prefer the status if it was provided
        if status is not None:
            if isinstance(status, (text_type, bytes, bytearray)):
                rv.status = status
            else:
                rv.status_code = status

        # extend existing headers with provided headers
        if headers:
            rv.headers.extend(headers)

        return rv

    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        response = self._middleware_chain(request)
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def run(self, host="0.0.0.0", port=5000):
        is_debug = True if self.config.get("debug", False) else False
        run_simple(host, port, self, use_debugger=is_debug, use_reloader=False)



from werkzeug.wrappers import Response
from json import dumps
from importlib import import_module


text_type = str


def jsonify(*args, **kwargs):
    indent = None
    separators = (",", ":")

    if args and kwargs:
        raise TypeError("jsonify() behavior undefined when passed both args and kwargs")
    elif len(args) == 1:  # single args are passed directly to dumps()
        data = args[0]
    else:
        data = args or kwargs

    return Response(
        dumps(data, indent=indent, separators=separators) + "\n",
        mimetype="application/json",
    )


def reraise(tp, value, tb=None):
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value


def _endpoint_from_view_func(view_func):
    """Internal helper that returns the default endpoint for a given
    function.  This always is the function name.
    """
    assert view_func is not None, "expected view func if endpoint is not provided."
    return view_func.__name__


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (module_path, class_name)) from err


def load_config(kwargs: dict)->dict:
    config = {}
    for k, v in kwargs.get("config", {}).items():
        if isinstance(k, str):
            config.update({
                k: v,
                k.lower(): v,
            })
        else:
            config.update({k: v})
    return config



# -*- coding:utf-8 -*-

import os
import sys
from distutils.core import setup
from setuptools import find_packages

os.chdir(os.path.dirname(sys.argv[0]) or ".")
here = os.path.abspath(os.path.dirname(__file__))

setup_args = dict(
    name='madara',
    version='0.0.4',
    description='A simple web framework.',
    long_description="Madara is a web framework inspire by [flask](https://github.com/pallets/flask).",
    long_description_content_type="text/markdown",
    author='arvin',
    license='MIT',
    url='https://github.com/Arvintian/madara',
    author_email='arvintian8@gamil.com',
    packages=find_packages(),
    include_package_data=True
)

if 'setuptools' in sys.modules:
    setup_args['zip_safe'] = False
    setup_args['install_requires'] = install_requires = []

    with open('requirements.txt') as f:
        for line in f.readlines():
            req = line.strip()
            if not req or req.startswith('#') or '://' in req:
                continue
            install_requires.append(req)


def main():
    setup(**setup_args)


if __name__ == "__main__":
    main()



from madara.utils import _endpoint_from_view_func


class BlueprintSetupState(object):

    def __init__(self, blueprint, app, options):
        self.app = app

        self.blueprint = blueprint

        self.options = options

        url_prefix = self.options.get("url_prefix")
        if url_prefix is None:
            url_prefix = self.blueprint.url_prefix

        self.url_prefix = url_prefix

    def add_url_rule(self, pattern, endpoint=None, view_func=None, **options):
        if self.url_prefix is not None:
            if pattern:
                pattern = "/".join((self.url_prefix.rstrip("/"), pattern.lstrip("/")))
            else:
                pattern = self.url_prefix

        if endpoint is None:
            endpoint = _endpoint_from_view_func(view_func)

        self.app.add_url_rule(
            pattern,
            "%s.%s" % (self.blueprint.name, endpoint),
            view_func,
            **options
        )


class Blueprint(object):

    def __init__(self, name, url_prefix=None):
        self.name = name
        self.url_prefix = url_prefix
        self.deferred_functions = []

    def record(self, func):
        """Registers a function that is called when the blueprint is
        registered on the application.
        """
        self.deferred_functions.append(func)

    def make_setup_state(self, app, options):
        return BlueprintSetupState(self, app, options)

    def register(self, app, options):
        state = self.make_setup_state(app, options)
        for deferred in self.deferred_functions:
            deferred(state)

    def route(self, pattern, **options):

        def decorator(func):
            endpoint = options.pop("endpoint", func.__name__)
            self.add_url_rule(pattern, endpoint, func, **options)
            return func

        return decorator

    def add_url_rule(self, pattern, endpoint=None, view_func=None, **options):

        if endpoint:
            assert "." not in endpoint, "Blueprint endpoints should not contain dots"
        if view_func and hasattr(view_func, "__name__"):
            assert (
                "." not in view_func.__name__
            ), "Blueprint view function name should not contain dots"
        self.record(lambda s: s.add_url_rule(pattern, endpoint, view_func, **options))
